#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "dosfd_img.h"


#define MAP_FILE			"MAP.DAT"
#define MAP_SIZE			(0x1000)
#define MAP_OFFSET_PLUTO	(0x27000)

#define PLUTO_PC98_MAP_FILE	"u2pc98pluto.map"


static unsigned char map_data[0x29000];
static unsigned char map_pluto[MAP_SIZE];


static void usage(void);
static char *create_write_name(const char *name);
static int read_map_file(const char *filename);
static int confirm(const char *msg);


int main(int argc, char **argv)
{
	int result = 1;
	int num_cmd_params = 1;
	const char *img_file = "U2_Ultima.fdi";
	const char *out_file = NULL;
	const char *map_file = PLUTO_PC98_MAP_FILE;
	char *out_file_buf = NULL;

	int opt_overwrite = 0;

	struct dosfd_image_t image;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "-o") == 0 || strcmp(*argv, "--write") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: no output filename\n");
					return 1;
				}
				out_file = *argv;
			} else if (strcmp(*argv, "-m") == 0 || strcmp(*argv, "--map") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: no map filename\n");
					return 1;
				}
				map_file = *argv;
#if 0
			} else if (strcmp(*argv, "--overwrite") == 0) {
				opt_overwrite = 1;
#endif
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ileegal parameter '%s'\n", *argv);
				return 1;
			}
			img_file = *argv;
			num_cmd_params--;
		}
	}

	if (out_file == NULL) {
		if ((out_file_buf = create_write_name(img_file)) == NULL) {
			printf("ERROR: can't create filename\n");
			goto ERROR_EXIT;
		}
		out_file = out_file_buf;
	}

	if (read_map_file(map_file) < 0) {
		printf("ERROR: read map file '%s'\n", map_file);
		goto ERROR_EXIT;
	}

	if (read_dosfd_image(&image, img_file, FDI_OFFSET) < 0) {
		printf("ERROR: read disk image '%s'\n", img_file);
		goto ERROR_EXIT;
	}

	if (read_dosfd_image_file(&image, MAP_FILE, map_data, sizeof(map_data), 1) < 0) {
		printf("ERROR: read disk image '%s/%s'\n", img_file, MAP_FILE);
		goto ERROR_EXIT;
	}

	memcpy(map_data + MAP_OFFSET_PLUTO, map_pluto, MAP_SIZE);

	if (write_dosfd_image_file(&image, MAP_FILE, map_data, sizeof(map_data), 1) < 0) {
		printf("ERROR: write disk image '%s/%s'\n", img_file, MAP_FILE);
		goto ERROR_EXIT;		
	}

	if (opt_overwrite == 0) {
		struct stat st;
		if (stat(out_file, &st) == 0) {
			printf("already exists '%s'\n", out_file);
			if (confirm("overwrite ?") <= 0) {
				goto ERROR_EXIT;
			}
		}
	}

	if (write_dosfd_image(&image, out_file) < 0) {
		printf("ERROR: write disk image '%s'\n", img_file);
		goto ERROR_EXIT;
	}

	free_dosfd_image(&image, 0);

	result = 0;

ERROR_EXIT:
	if (out_file_buf != NULL) {
		free(out_file_buf);
	}

	return result;
}

static void usage(void)
{
	printf("usage: u2pc98pluto [options] [DISK-IMG]\n");
	printf("option: -o|--write OUTPUT-IMG   specify output file name\n");
	printf("        -m|--map PLUTO-MAP      specify map file name\n");
#if 0
	printf("           --overwrite          overwrite output files\n");
#endif
	printf("        -?|--help               print usage\n");
	printf("\n");
	printf("read disk image from 'U2_Ultima.fdi' if not specified\n");
	printf("read map data from 'u2pc98pluto.map' if not specified\n");
}

static char *create_write_name(const char *name)
{
	int baselen;
	const char *ext;
	int len = strlen(name) + 5;
	char *new_name = (char *)malloc(len);
	if (new_name == NULL) {
		printf("memory exhaust!\n");
		return NULL;
	}

	if ((ext = strrchr(name, '.')) == NULL) {
		baselen = strlen(name);
		ext = "";
	} else {
		baselen = ext - name;
	}

	strncpy(new_name, name, baselen);
	strcpy(new_name + baselen, ".NEW");
	strcpy(new_name + baselen + 4, ext);

	return new_name;
}

static int read_map_file(const char *filename)
{
	int result = -1;
	FILE *fp = NULL;

	if ((fp = fopen(filename, "rb")) == NULL) {
		printf("ERROR: open error: %s\n", filename);
		goto ERROR_EXIT;
	}

	if (fread(map_pluto, sizeof(map_pluto), 1, fp) != 1) {
		printf("ERROR: read error: %s\n", filename);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
